﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;
using System.Reflection;
using System.Media;
using System.Threading;

namespace MIDI_Synthesizer
{
    public partial class Form1 : Form
    {
        FileStream fs;

        // List that stores all the midievents of a single midi file
        List<midiEvent> midiEventList = new List<midiEvent>();

        int threadCount = 1;

        // Threads to play upto 24 tunes at a same time
        Thread t1;
        Thread t2;
        Thread t3;
        Thread t4;
        Thread t5;
        Thread t6;
        Thread t7;
        Thread t8;
        Thread t9;
        Thread t10;
        Thread t11;
        Thread t12;
        Thread t13;
        Thread t14;
        Thread t15;
        Thread t16;
        Thread t17;
        Thread t18;
        Thread t19;
        Thread t20;
        Thread t21;
        Thread t22;
        Thread t23;
        Thread t24;


        const int format = 0;
        const int numOfTracks = 1;
        const int division = 192;       // Ticks per beat
        const int on = 221;             // equivalent of "on"
        const int of = 213;             // equivalent of "of"
        const int velocity = 127;       // Velocity feature is not implemented
        const int frameByteSize = 16;   // Size of the frame sent by the microcontroller for each event
        const int tempo = 500000;       // uS per Beat...BPM = MICROSECONDS_PER_MINUTE / Tempo..now BPM = 120

        // Standard piano note frequencies
        int[] noteFrequencies = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 27, 29, 31, 
                                  32, 34, 36, 38, 41, 43, 46, 49, 52, 55, 58, 61,
                                  65, 69, 73, 77, 82, 87, 92, 98, 104, 110, 116, 123, 
                                  131, 138, 147, 155, 165, 175, 185, 196, 207, 220, 233, 247, 
                                  261, 277, 293, 311, 329, 349, 370, 392, 415, 440, 466, 494, 
                                  523, 554, 587, 622, 659, 698, 740, 784, 830, 880, 932, 988, 
                                  1046, 1109, 1175, 1244, 1318, 1397, 1480, 1568, 1661, 1760, 1864, 1979, 
                                  2093, 2217, 2349, 2489, 2637, 2794, 2960, 3136, 3322, 3520, 3729, 3851, 
                                  4186, 4434, 4698, 4978, 5274, 5587, 5919, 6271, 6644, 7040, 7458, 7902, 
                                  8372, 8869, 9397, 9956, 10548, 11175, 11839, 12543, 13289, 14080, 14917, 15804, 
                                  16744, 17739, 18794, 19912, 21096, 22350, 23679, 25087};

        byte patch = 1;                 // Instrument selection
        int track1Length = 0;

        double BPM;
        double ticksPerMinute;
        double tickDurationMS;

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            saveFileDialog1.Filter = "MIDI files (*.mid)|*.mid";
            saveFileDialog1.Title = "Save MIDI file as";
            saveFileDialog1.RestoreDirectory = true;

            // delta ticks calculation
            BPM = (60000000 / tempo);
            ticksPerMinute = (BPM * division);
            tickDurationMS = (60000 / ticksPerMinute);

            // Initialize the form components
            string[] ports = SerialPort.GetPortNames();
            foreach (string str in ports)
            {
                serialportName.Items.Add(str);
            }

            serialportName.SelectedIndex = serialportName.Items.Count - 1;
            baudRate.SelectedText = "9600";
            instrument.SelectedIndex = 0;
            checkBox1.Checked = true;
        }

        private void button1_Click(object sender, EventArgs e)
        {
            closeFileAndComPort();

            try
            {
                serialPort1 = new System.IO.Ports.SerialPort(serialportName.Text, int.Parse(baudRate.Text));
                serialPort1.Open();
            }
            catch (Exception exp)
            {
                MessageBox.Show("Unable to open the sepcified COM port","Error",MessageBoxButtons.OK,MessageBoxIcon.Error);
            }

            if (serialPort1.IsOpen)
            {
                serialPort1.DiscardInBuffer();

                if (saveFileDialog1.ShowDialog() == DialogResult.OK)
                {
                    fs = new FileStream(saveFileDialog1.FileName, FileMode.Create, FileAccess.Write);

                    if (fs != null)
                    {
                        patch = (byte)(instrument.SelectedIndex + 1);
                        midiHeaderWriter();

                        timer1.Start();

                        serialPort1.Write("a");     // works like start byte for uC
                    }
                    else
                        MessageBox.Show("Invalid selection", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                else
                {
                    MessageBox.Show("Invalid selection", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }

            }
        }

        // Closes the opened COM port and saves and closes the MIDI file
        public void closeFileAndComPort()
        {
            timer1.Stop();

            if (serialPort1.IsOpen)
            {
                serialPort1.DiscardInBuffer();
                serialPort1.Close();
            }

            if (fs != null)
            {
                decodeEvents();

                fs.WriteByte(0);
                fs.WriteByte(0xff);
                fs.WriteByte(0x2f);
                fs.WriteByte(0);
                track1Length += 4;
                fs.Seek(18, SeekOrigin.Begin);              // update the length info of the file
                char[] temp1 = numToChar(track1Length, 4);
                fs.WriteByte((byte)temp1[0]);
                fs.WriteByte((byte)temp1[1]);
                fs.WriteByte((byte)temp1[2]);
                fs.WriteByte((byte)temp1[3]);
                fs.Close();
            }

            track1Length = 0;
            midiEventList.Clear();
        }

        private void midiHeaderWriter()
        {
            fs.WriteByte((byte)'M');
            fs.WriteByte((byte)'T');
            fs.WriteByte((byte)'h');
            fs.WriteByte((byte)'d');
            // length of header
            fs.WriteByte((byte)0);
            fs.WriteByte((byte)0);
            fs.WriteByte((byte)0);
            fs.WriteByte((byte)6);
            // format
            fs.WriteByte((byte)0);
            fs.WriteByte((byte)format);
            // number of tracks
            char[] temp1 = numToChar(numOfTracks, 2);
            fs.WriteByte((byte)temp1[0]);
            fs.WriteByte((byte)temp1[1]);
            // division
            char[] temp2 = numToChar(division, 2);
            fs.WriteByte((byte)temp2[0]);
            fs.WriteByte((byte)temp2[1]);

            midiTrackChunkStart();

        }

        private void midiTrackChunkStart()
        {
            fs.WriteByte((byte)'M');
            fs.WriteByte((byte)'T');
            fs.WriteByte((byte)'r');
            fs.WriteByte((byte)'k');
            // track length
            char[] temp1 = numToChar(track1Length, 4);
            fs.WriteByte((byte)temp1[0]);
            fs.WriteByte((byte)temp1[1]);
            fs.WriteByte((byte)temp1[2]);
            fs.WriteByte((byte)temp1[3]);
            
            // program or patch change
            fs.WriteByte(0);
            fs.WriteByte(0xc0);
            fs.WriteByte(patch);

            track1Length += 3;

            // control change
            fs.WriteByte(0);
            fs.WriteByte(0xb0);
            fs.WriteByte(0xa0);
            fs.WriteByte(0x40);

            track1Length += 4;

            // set tempo
            fs.WriteByte(0);
            fs.WriteByte(0xff);
            fs.WriteByte(0x51);
            fs.WriteByte(0x03);
            char[] temp2 = numToChar(tempo, 3);
            fs.WriteByte((byte)temp2[0]);
            fs.WriteByte((byte)temp2[1]);
            fs.WriteByte((byte)temp2[2]);


            track1Length += 7;
        }

        // Represents a given number in the number of specifiec bytes
        private char[] numToChar(int num, int numOfBytes)
        {
            char[] temp = new char[numOfBytes];

            int shift = numOfBytes * 8;
            int mask = (0xff << (shift - 1));

            for (int i = 0; i < numOfBytes; i++)
            {
                shift -= 8;
                temp[i] = (char)(num >> shift);
                num &= (~mask);
                mask >>= 8;
            }

            return temp;
        }

        private void Form1_FormClosing(object sender, FormClosingEventArgs e)
        {
            closeFileAndComPort();
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            int i = serialPort1.BytesToRead;

            i /= frameByteSize;     // Read only complete frames

            for (; i > 0; i--)
            {
                getEvents();
            }
        }

        public void getEvents()
        {
            int onOff = 0;
            int note = 0;
            int time = 0;
            byte channel = 0;

            onOff = serialPort1.ReadByte();
            onOff += serialPort1.ReadByte();

            if ((onOff == on) || (onOff == of))
            {
                channel = (byte)(serialPort1.ReadByte() - 48);

                for (int i = 0; i < 3; i++)
                {
                    note += (int)((serialPort1.ReadByte() - 48) * Math.Pow(10, (2 - i)));
                }

                if(onOff == on)
                    playBack(channel, note);

                for (int i = 0; i < 10; i++)
                {
                    time += (int)((serialPort1.ReadByte() - 48) * Math.Pow(10, (9 - i)));
                }

                midiEventList.Add(new midiEvent(onOff,channel,note,time));
            }
        }
                        
        public void decodeEvents()
        {
            int deltaTicks = 0;

            byte[] deltaTime;

            while (midiEventList.Count != 0)
	        {
		        if ((midiEventList[0].onOff == on) || (midiEventList[0].onOff == of))
                    {                
                        deltaTicks = (int)(((double)(midiEventList[0].time / 1000) / tickDurationMS));

                        deltaTime = variableLengthEncoder(deltaTicks);
                        
                        for (int i = 0; i < deltaTime.Length; i++)
                        {
                            fs.WriteByte(deltaTime[i]);
                            track1Length++;
                        }

                        if (midiEventList[0].onOff == on)
                            fs.WriteByte((byte)(0x90 | midiEventList[0].channel));
                        else if (midiEventList[0].onOff == of)
                            fs.WriteByte((byte)(0x80 | midiEventList[0].channel));

                        fs.WriteByte((byte)midiEventList[0].note);
                        fs.WriteByte((byte)velocity);

                        track1Length += 3;
                    }

                    midiEventList.RemoveAt(0); // Remove the event that was written to the file from the event list
	        }
        }

        // Encodes a given deltatime in a variable length standard MIDI deltatime format
        public byte[] variableLengthEncoder(int num)
        {
            // lower 7 bits represent the data, MSB represents whether the current byte is the last byte or not. And MIDI follows big endian method

            int number = num;
            int lengthOfArray = (int)((Math.Log(num, 2) / 7.0)+1);

            byte[] encodedResult = new byte[lengthOfArray];
            int index = lengthOfArray - 1;

            encodedResult[index--] = (byte)(num & 127);
            num >>= 7;

            while (num != 0)
            {
                encodedResult[index] = (byte)(num & 127);
                encodedResult[index--] |= 128;
                num >>= 7;
            }

            return encodedResult;
        }
        
        // Plays the current note on a new thread. It can play upto 24 notes simultaneously
        public void playBack(int octave, int note)
        {
            int noteIndex = ((octave * 12) + note);

            switch (threadCount)
            {
                case 1: t1 = new Thread(play);
                    t1.Start(noteFrequencies[noteIndex]);
                    break;

                case 2: t2 = new Thread(play);
                    t2.Start(noteFrequencies[noteIndex]);
                    break;

                case 3: t3 = new Thread(play);
                    t3.Start(noteFrequencies[noteIndex]);
                    break;

                case 4: t4 = new Thread(play);
                    t4.Start(noteFrequencies[noteIndex]);
                    break;

                case 5: t5 = new Thread(play);
                    t5.Start(noteFrequencies[noteIndex]);
                    break;

                case 6: t6 = new Thread(play);
                    t6.Start(noteFrequencies[noteIndex]);
                    break;

                case 7: t7 = new Thread(play);
                    t7.Start(noteFrequencies[noteIndex]);
                    break;

                case 8: t8 = new Thread(play);
                    t8.Start(noteFrequencies[noteIndex]);
                    break;

                case 9: t9 = new Thread(play);
                    t9.Start(noteFrequencies[noteIndex]);
                    break;

                case 10: t10 = new Thread(play);
                    t10.Start(noteFrequencies[noteIndex]);
                    break;

                case 11: t11 = new Thread(play);
                    t11.Start(noteFrequencies[noteIndex]);
                    break;

                case 12: t12 = new Thread(play);
                    t12.Start(noteFrequencies[noteIndex]);
                    break;

                case 13: t13 = new Thread(play);
                    t13.Start(noteFrequencies[noteIndex]);
                    break;

                case 14: t14 = new Thread(play);
                    t14.Start(noteFrequencies[noteIndex]);
                    break;

                case 15: t15 = new Thread(play);
                    t15.Start(noteFrequencies[noteIndex]);
                    break;

                case 16: t16 = new Thread(play);
                    t16.Start(noteFrequencies[noteIndex]);
                    break;

                case 17: t17 = new Thread(play);
                    t17.Start(noteFrequencies[noteIndex]);
                    break;

                case 18: t18 = new Thread(play);
                    t18.Start(noteFrequencies[noteIndex]);
                    break;

                case 19: t19 = new Thread(play);
                    t19.Start(noteFrequencies[noteIndex]);
                    break;

                case 20: t20 = new Thread(play);
                    t20.Start(noteFrequencies[noteIndex]);
                    break;

                case 21: t21 = new Thread(play);
                    t21.Start(noteFrequencies[noteIndex]);
                    break;

                case 22: t22 = new Thread(play);
                    t22.Start(noteFrequencies[noteIndex]);
                    break;

                case 23: t23 = new Thread(play);
                    t23.Start(noteFrequencies[noteIndex]);
                    break;

                case 24: t24 = new Thread(play);
                    t24.Start(noteFrequencies[noteIndex]);
                    break;
            
                default: break;
            }
            threadCount++;
            if (threadCount > 24)
                threadCount = 1;
            return;
                    
        }

        // Plays a sine wave of the given frequency
        public static void play(object freq)
        {
            int frequency = (int)freq;
            if(frequency > 37 && frequency < 32767)
                Console.Beep(frequency, 300);
        }
    }

    public class midiEvent
    {
        public int onOff;
        public byte channel;
        public int note;
        public int time;

        public midiEvent(int cmd, byte ch, int noteNumber, int time)
        {
            this.onOff = cmd;
            this.channel = ch;
            this.note = noteNumber;
            this.time = time;
        }
    };
}
